''' Translation functions that take Cells/Devices in different types of languages
    and convert them to Cells/Devices in other languages

    Supported: phidl, pya (GSI version), klayout.db (standalone version)
    Not yet supported: nazca, IPKISS, gdspy, other suggestions?
'''
import os

default_phidl_portlayer = 41
do_write_ports = True

def celltype_to_write_function(celltype):
    ''' Takes a class that represents a layout Cell and gives a function that writes its geometry to disc
        This is typically an unbound method of the class.
        (there is always some version of this, although some languages don't explicitly call it Cell)

        This uses try-excepted on-the-fly importing because not everybody is going to have every language installed.
        But if they don't, then their celltype is not going to be from that language.
    '''
    def not_supported_error(language_name):
        raise NotImplementedError(('The translator does not yet support {}.\n'.format(language_name)
                                    + 'Supported languages are: pya, phidl\n'
                                    + 'Future support will include: gdspy, nazca, ipkiss'))

    if type(celltype) is not type:
        celltype = type(celltype)

    try: import pya  # ok I know we already imported this, but in the future it can also be on-the-fly (i.e. converting phidl to nazca)
    except ImportError: pass
    else:
        if issubclass(celltype, pya.Cell):
            return pya.Cell.write
        elif issubclass(celltype, pya.Layout):
            return pya.Layout.write

    try: import phidl
    except ImportError: pass
    else:
        if issubclass(celltype, (phidl.Device, phidl.device_layout.DeviceReference)):
            if do_write_ports:
                def write_with_ports(device, filename, *args, port_layer=None, **kwargs):
                    if port_layer is None:
                        port_layer = default_phidl_portlayer
                    # If its a reference, use its parent
                    if issubclass(celltype, phidl.device_layout.DeviceReference):
                        device = device.parent
                    # Try to convert to geometric ports
                    try:
                        port2geom = phidl.geometry.ports_to_geometry
                    except AttributeError:  # it is an older version of phidl
                        pass
                    else:
                        device = port2geom(device, layer=port_layer)
                    # The actual write
                    device.write_gds(filename, *args, **kwargs)
                return write_with_ports
            else:
                def write_parent(device, filename, *args, **kwargs):
                    # If its a reference, use its parent
                    if issubclass(celltype, phidl.device_layout.DeviceReference):
                        device = device.parent
                    device.write_gds(filename, *args, **kwargs)
                return write_parent

    # try: import gdspy
    # except ImportError: pass
    # else:
    #     not_supported_error('gdspy')

    # try: import nazca
    # except ImportError: pass
    # else:
    #     not_supported_error('nazca')

    # try: import ipkiss
    # except ImportError: pass
    # else:
    #     not_supported_error('ipkiss')

    raise TypeError('celltype: {} is not recognized as a layout cell object'.format(celltype.__name__))


def any_write(cell, *args, **kwargs):
    write = celltype_to_write_function(cell)
    return write(cell, *args, **kwargs)


def celltype_to_read_function(celltype):
    ''' Takes a class that represents a layout Cell and gives its class' read method.
    '''
    def not_supported_error(language_name):
        raise NotImplementedError(('The translator does not yet support {}.\n'.format(language_name)
                                    + 'Supported languages are: pya, phidl\n'
                                    + 'Future support will include: gdspy, nazca, ipkiss'))

    if type(celltype) is not type:
        celltype = type(celltype)

    try: import pya  # ok I know we already imported this, but in the future it can also be on-the-fly (i.e. converting phidl to nazca)
    except ImportError: pass
    else:
        if issubclass(celltype, pya.Cell):
            def pyaCell_reader(pya_cell, filename, *args, **kwargs):
                templayout = pya.Layout()
                templayout.read(filename)
                tempcell = templayout.top_cell()
                # Transfer the geometry of the imported cell to the one specified
                pya_cell.name = tempcell.name
                pya_cell.copy_tree(tempcell)
                return pya_cell
            return pyaCell_reader

    try: import phidl
    except ImportError: pass
    else:
        if issubclass(celltype, phidl.Device):
            def phidlDevice_reader(phidl_device, filename, *args, port_layer=None, **kwargs):
                # phidl_device is not really used. It is just there to determine type.
                if 'cellname' in kwargs:
                    cellname = kwargs.pop('cellname')
                else:
                    #### hacks, because sometimes pya saves an extra topcell called $$$CONTEXT_INFO$$$
                    from gdspy import GdsLibrary
                    gdsii_lib = GdsLibrary()
                    gdsii_lib.read_gds(filename)
                    top_level_cells = gdsii_lib.top_level()
                    if len(top_level_cells) == 1:
                        cellname = top_level_cells[0].name
                    elif len(top_level_cells) == 2:
                        for tc in top_level_cells:
                            if tc.name != '$$$CONTEXT_INFO$$$':
                                cellname = tc.name
                    else:
                        raise ValueError('There are multiple top level cells: {}.\n Please specify with the cellname argument.'.format(top_level_cells))
                    #### end hacks

                # main read function
                tempdevice = phidl.geometry.import_gds(filename, *args, cellname=cellname, **kwargs)

                # check for port geometry
                try:
                    wop = phidl.geometry.with_object_ports
                except AttributeError:
                    pass
                else:
                    if port_layer is None:
                        port_layer = default_phidl_portlayer
                    tempdevice = wop(tempdevice, layer=port_layer)
                # copy over from temporary device
                phidl_device.polygons = tempdevice.polygons
                phidl_device.references = tempdevice.references
                phidl_device.ports = tempdevice.ports
                phidl_device.labels = tempdevice.labels
                phidl_device.name = tempdevice.name
                return phidl_device
            return phidlDevice_reader

    # try: import gdspy
    # except ImportError: pass
    # else:
    #     not_supported_error('gdspy')

    # try: import nazca
    # except ImportError: pass
    # else:
    #     not_supported_error('nazca')

    # try: import ipkiss
    # except ImportError: pass
    # else:
    #     not_supported_error('ipkiss')

    raise TypeError('celltype: {} is not recognized as a layout cell object'.format(celltype.__name__))


def any_read(cell, *args, **kwargs):
    read = celltype_to_read_function(cell)
    return read(cell, *args, **kwargs)


def anyCell_to_anyCell(initial_cell, final_cell):
    ''' Transfers the geometry of some initial_cell into another format.
        This initial_cell can be any type of layout object in any supported language.

        initial_cell must provide a way to write its geometry to a file and have only one top cell.
        This function will figure out what that way is, based on the class of initial_cell.

        final_cell must provide a way to read geometry in from a file.

        The supported types and their mapping to write methods are contained in celltype_to_write_function and celltype_to_read_function.
    '''
    global do_write_ports
    do_write_ports_orig = do_write_ports
    do_write_ports = True
    tempfile = os.path.expanduser('~/temp_cellTranslation.gds')
    any_write(initial_cell, tempfile)
    new_cell = any_read(final_cell, tempfile)
    os.remove(tempfile)
    do_write_ports = do_write_ports_orig

    # Transfer other data (ports, metadata, CML files, etc.)
    pass  # TODO

    return new_cell



''' Safe versions of the GUI members for window and active layout.
    These use patch_environment so they will work fine in GUI/GSI and at least not error in standalone.
'''
import pya
from lygadgets.technology import Technology


def gui_window():
    patch_environment()  # makes sure the Application attribute gets spoofed into the standalone
    import pya
    return pya.Application.instance().main_window()


def gui_view():
    lv = gui_window().current_view()
    if lv is None:
        raise UserWarning("No view selected. Make sure you have an open layout.")
    return lv


def gui_active_layout():
    ly = gui_view().active_cellview().layout()
    if ly is None:
        raise UserWarning("No layout. Make sure you have an open layout.")
    return ly


def gui_active_cell():
    cell = gui_view().active_cellview().cell
    if cell is None:
        raise UserWarning("No cell. Make sure you have an open layout.")
    return cell


def gui_active_technology_name():
    return gui_window().initial_technology  # gets the technology from the selection menu


def gui_active_technology():
    technology = gui_active_technology_name()
    tech_obj = Technology.technology_by_name(technology)
    return tech_obj



'''
    lygadgets.Technology serves as a drop-in for pya.Technology in the pymod standalone.

    The difference is that the lygadgets version loads all the installed technologies in salt.
    This is not a bug. The intention of pymod:pya.Technology is to isolate from the state of a user's machine,
    but some packages are not associated with a particular technology, so they have to get it from salt.

    Nothing happens on import: Autoload from salt is triggered when you ask for either
    - Technology.technology_names, or
    - Technology.technology_by_name, or
    - Technology.has_technology

    lygadgets.Technology also offers a new method: register_lyt
    which takes the .lyt file, turns it into a Technology object (eventually returns), and adds to the class registry
'''
from lygadgets import klayout_home, pya
import xmltodict
import os

# default active is tricky and a bad idea if it just for CL convenience
# perhaps lygadgets has an idea of active technology
# this would have to be influenced by GUI switches, but those are not global

try:
    import pya
except ImportError:
    print('Did not find pya. You will not be able to use lygadgets.Technology')
    class Technology(object):
        pass
else:
    class Technology(pya.Technology):
        ''' '''
        _salt_loaded = False

        @classmethod
        def _register_pyatech(cls, pya_tech):
            if cls.has_technology(pya_tech.name):
                new_tech = cls.technology_by_name(pya_tech.name)
            else:
                new_tech = cls.create_technology(pya_tech.name)

            # Registering new technology to klayout's database
            new_tech.assign(pya_tech)
            return new_tech

        @classmethod
        def register_lyt(cls, lyt_filename):
            pya_tech = _load_pya_tech(lyt_filename)
            cls._register_pyatech(pya_tech)
            return pya_tech

        @classmethod
        def crawl_for_technologies(cls, path):
            for root, dirnames, filenames in os.walk(path, followlinks=True):
                for fn in filenames:
                    if fn.endswith('.lyt'):
                        cls.register_lyt(os.path.join(root, fn))

        @classmethod
        def _load_salt(cls):
            ''' Crawls through klayout_home directory looking for .lyt files.
                Registers them with the class, returns nothing
            '''
            if cls._salt_loaded:
                return None
            cls._salt_loaded = True
            if os.path.isdir(klayout_home()):
                cls.crawl_for_technologies(os.path.join(klayout_home(), 'salt'))
                cls.crawl_for_technologies(os.path.join(klayout_home(), 'tech'))


        # These override methods of pya.Technology

        @classmethod
        def technology_names(cls):
            ''' Equivalent behavior to pya.Technology.technology_names() in klayout's GSI '''
            cls._load_salt()
            return super().technology_names()

        @classmethod
        def technology_by_name(cls, tech_name):
            ''' Equivalent behavior to pya.Technology.technology_by_name() in klayout's GSI '''
            cls._load_salt()
            return super().technology_by_name(tech_name)

        @classmethod
        def has_technology(cls, tech_name):
            ''' Equivalent behavior to pya.Technology.has_technology() in klayout's GSI '''
            cls._load_salt()
            return super().has_technology(tech_name)

        # End of overrides.


    def _load_pya_tech(lyt_filename):
        ''' Parses the .lyt which is in xml format.
            Returns the new Technology object. Does not register it to the Technology class
        '''
        # workaround while https://github.com/klayoutmatthias/klayout/pull/215 is not solved
        absolute_filepath = os.path.realpath(os.path.expanduser(lyt_filename))
        with open(absolute_filepath, 'r') as file:
            lyt_xml = file.read()
        pya_tech = Technology.technology_from_xml(lyt_xml)
        pya_tech.default_base_path = os.path.dirname(absolute_filepath)
        # end of workaround
        return pya_tech

def klayout_last_open_technology():
    # use this to pick out a starting "active" technology
    if os.path.isdir(klayout_home()):
        rc_file = os.path.join(klayout_home(), 'klayoutrc')
        if os.path.isfile(rc_file):
            with open(rc_file, 'r') as file:
                rc_dict = xmltodict.parse(file.read(), process_namespaces=True)
            return rc_dict['config']['initial-technology']
    return ''



import xmltodict

def xml_to_dict(xml_content):
    # to be deprecated
    return xmltodict.parse(xml_content, process_namespaces=True)


def lyp_to_layerlist(lyp_file):
    ''' Now works with multiple tabs. It ignores the tab names '''
    with open(lyp_file, 'r') as fx:
        lyp_dict = xmltodict.parse(fx.read(), process_namespaces=True)
    all_layers = []
    if 'layer-properties-tabs' in lyp_dict.keys():
        all_layer_tabs = lyp_dict['layer-properties-tabs']['layer-properties']
    else:
        all_layer_tabs = [lyp_dict['layer-properties']]

    for tab in all_layer_tabs:
        tab_layers = tab['properties']
        if not isinstance(tab_layers, list):
            tab_layers = [tab_layers]
        all_layers.extend(tab_layers)
    return all_layers



''' This module converts pcells of any type into KLayout-style PCells

    All of these things are not specific to any technology.
    Eventually, this module could be part of an independent PDK specification package.

    It is a bad idea to have this "python" directory permanently because
    this is a technology specification package, not a functional codebase.
'''
# Todo: handle TypeList and TypeLayer arguments
# Going to be difficult to handle: arguments that are other phidl.Device objects or functions

from lygadgets import pya
import os
from inspect import signature
from lygadgets import anyCell_to_anyCell

def cellname_to_kwargs(cellname):
    ''' Converts the naming convention for parameter-based cell naming back into a dict of kwargs '''
    # this function has not been tested yet
    func_kwargs = dict()
    func_name, paramstr = cellname.split('__')  # double underscore has been sanitized
    for oneparamstr in paramstr.split('_'):  # todo: unsanitize _ and =
        name, val = oneparamstr.split('=')
        try:
            val = int(val)
        except ValueError:
            try:
                val = float(val)
            except ValueError:
                pass
        func_kwargs[name] = val
    return func_kwargs


def pytype_to_PCelltype(arg_type):
    ''' Maps builtin python types to the corresponding types of the PCellDeclarationHelper class
    '''
    # this has been tested only with floats so far
    if issubclass(arg_type, int):
        return pya.PCellDeclarationHelper.TypeInt
    if issubclass(arg_type, float):
        return pya.PCellDeclarationHelper.TypeDouble
    if issubclass(arg_type, (list, tuple)):
        return pya.PCellDeclarationHelper.TypeList
    if issubclass(arg_type, str):
        return pya.PCellDeclarationHelper.TypeString
    if issubclass(arg_type, bool):
        return pya.PCellDeclarationHelper.TypeBoolean
    if issubclass(arg_type, NoneType):
        return pya.PCellDeclarationHelper.TypeNone


def my_argspec(function):
    ''' Extracts the arguments and keyword arguments from any callable object.
        The dictionary of kwargs is returned with values reflecting their default arguments
    '''
    func_signature = signature(function)
    args = list()
    kwargs = dict()
    for key, param in func_signature.parameters.items():
        kwargs[key] = param.default
    return args, kwargs


# protect against missing pya
if pya is None:
    class WrappedPCell(object):
        def __init__(self, *args, **kwargs):
            raise AttributeError('WrappedPCell requires a working installation of klayout python package')
    class WrappedLibrary(object):
        def __init__(self, *args, **kwargs):
            raise AttributeError('WrappedLibrary requires a working installation of klayout python package')
else:
    class WrappedPCell(pya.PCellDeclarationHelper):
        ''' Wraps a non-klayout PCell as a klayout PCell.
            The pcell is here defined as a function with arguments.
            When produce_impl is called, the cell geometry is compiled and added to this instance's cell.

            The transfer of external pcell geometry is done through a temporary gds file.

            I think this is not specific to phidl implementation. It just needs some function.
            Oh wait, yes it is (barely) because of write_gds.
        '''
        generating_function = None

        def kwargs_to_params(self):
            ''' Extracts the arguments of the generating_function and registers them as params of this object
                If it is a keyword argument, the default type is detected and value added as default
            '''
            def kwarg_to_param(key, default=None):
                self.param(key, pytype_to_PCelltype(type(default)), key, default=default)
            args, kwargs = my_argspec(self.generating_function)
            for arg in args:
                kwarg_to_param(arg)
            for key, default in kwargs.items():
                kwarg_to_param(key, default)

        def params_to_kwargs(self):
            ''' Inverse of kwargs_to_params. Converts the parameters of this instance and their chosen values
                back into arguments that can be sent to the generating_function.
            '''
            # todo: handle non-keyword args
            all_args = list()
            all_kwargs = dict()
            for pdecl, pval in zip(self.get_parameters(), self.get_values()):
                all_kwargs[pdecl.name] = pval
            return all_args, all_kwargs

        def __init__(self, generating_function):
            self.generating_function = generating_function
            super().__init__()
            self.kwargs_to_params()

        def display_text_impl(self):
            ''' Produces a string that includes all of the parameters. This means it is unique for identical cells.
                We assume that the pcells are functional in the sense that identical parameters yield identical geometry.
                In this way, different pcell instances with the same arguments are correctly identified as the same cell.
            '''
            text = self.generating_function.__name__ + '_'
            for pdecl, pval in zip(self.get_parameters(), self.get_values()):
                # sanitize _'s and ='s
                if isinstance(pval, str):
                    pval.replace('_', '[_]')
                    pval.replace('=', '[=]')
                text += '_{}={}'.format(pdecl.name, pval)
            return text

        def produce_impl(self):
            ''' Creates a fixed cell instance based on the previously specified parameters
            '''
            # Produce the geometry
            args, kwargs = self.params_to_kwargs()
            phidl_Device = self.generating_function(*args, **kwargs)
            # Convert phidl.Device to pya.Cell - just geometry
            anyCell_to_anyCell(phidl_Device, self.cell)
            # Transfer other data (ports, metadata, CML files, etc.)
            pass  # TODO


    class WrappedLibrary(pya.Library):
        ''' An abstract library consisting of pya PCells that are
            based on function calls that possibly involve other languages (specified in all_funcs_to_wrap)
            The names of the PCells will end up being the same as the names of the functions.

            Initializing the library registers it in klayout's repository of libraries.

            To subclass this class, no extra methods are needed.
            Just override the class attributes.
        '''
        tech_name = None
        all_funcs_to_wrap = None
        description = None

        def __init__(self):
            if self.tech_name is None or self.all_funcs_to_wrap is None:
                raise NotImplementedError('WrappedLibrary must be subclassed.')

            print("Initializing '%s' Library." % self.tech_name)

            # Not doing fixed GDS in this Library

            # Create all the new klayout-format PCells
            for func in self.all_funcs_to_wrap:
                self.layout().register_pcell(func.__name__, WrappedPCell(func))  # generic version

            self.register(self.tech_name)

            if int(pya.Application.instance().version().split('.')[1]) > 24:
                # KLayout v0.25 introduced technology variable:
                self.technology = self.tech_name


# class WrappedKLayoutPCell(siepic.KLayoutPCell):
#     pass

def PCell_to_phidlfunction(pyaPCell):
    ''' Takes a pcell instance and turns its placement into a function with arguments.
        This function is returned.
        return_type is, for example, a phidl.Device
    '''
    import phidl
    all_args, all_kwargs = WrappedPCell.params_to_kwargs(pyaPCell)
    def pfunc(*args, **kwargs):
        for k in kwargs.keys():
            if k not in all_kwargs.keys():
                raise ValueError('Invalid argument. "{}" is not in {}'.format(k, all_kwargs.keys()))
        all_kwargs.update(kwargs)
        for k, v in all_kwargs.keys():
            setattr(pyaPCell, k, v)

        layout = pya.Layout()
        layout.place(pyaPCell.pcell())  # WRONG method. I forget what "place" is called

        new_Device = phidl.Device()
        anyCell_to_anyCell(layout, new_Device)
        return new_Device
    return pfunc





from __future__ import print_function
from lygadgets import pya, isGUI

if isGUI():
    main_window = pya.Application.instance().main_window()
    message = lambda msg: main_window.message(msg, 5000)
    message_loud = lambda msg: pya.MessageBox.info('Message for you', msg, pya.MessageBox.Ok)
else:
    PINK = '\033[95m'
    NORMAL_COLOR = '\033[0m'
    message = print
    message_loud = lambda msg: print(PINK + str(msg), NORMAL_COLOR)


import os
import sys
from sys import platform
from future.utils import with_metaclass

''' Detect whether we are in klayout's generic scripting interface
    Since standalone pya has no associated Application, try that

    Also, Determine what we will use for pya.
    Do not override it with the standalone in sys.modules yet
    This statement:
        from lygadgets import pya
    will not error even if you are not in klayout GSI and don't have the klayout.db standalone
'''
try:
    import pya
except ImportError:
    _isGSI = False
    # perhaps "pya" package didn't get there but you have the klayout package
    try:
        import klayout.db as pya
    except ImportError:
        _isStandalone = False
        pya = None
    else:
        _isStandalone = True
        print('Warning: You seem to be using an old version of klayout standalone')
        print('Warning: Try running "pip install --upgrade klayout"')
else:
    _isGSI = (pya.__package__ == '')  # If True, it implies that pya.Application exists
    _isStandalone = not _isGSI

isGSI = lambda: _isGSI
isStandalone = lambda: _isStandalone


''' Klayout can run as a window or in batch mode on command line.
    If it launches in batch (-b) or database-only (-zz) mode, then main_window is None.

    If however it runs in non-GUI mode (-z), it is not None; however,
    it has not been given a title yet. (see layApplication.cc).
    This is unstable, so we don't use that information.

    Do not expect this to work in (-z) mode.
'''
if isGSI():
    try:
        main = pya.Application.instance().main_window()
    except AttributeError:
        _isGUI = False
    else:
        _isGUI = (main is not None)
else:
    _isGUI = False
isGUI = lambda: _isGUI


def klayout_home():
    ''' Figure out the klayout configuration directory.
        Uses exactly the same logic as used in the klayout source.
    '''
    ly_home = os.environ.get('KLAYOUT_HOME', '')
    if ly_home == '':
        ly_home = os.path.join(os.path.expanduser('~'), 'KLayout' if is_windows() else '.klayout')
    if not os.path.exists(ly_home):
        print('The KLayout config directory was not found. KLayout might not be installed.')
        print('Creating it just in case that solves things')
        os.mkdir(ly_home)
        # raise FileNotFoundError('The KLayout config directory was not found. KLayout might not be installed.')
    return ly_home


def klayout_version():
    if isGUI():
        return pya.Application.instance().version()
    else:
        return '0.26.0'  # Hard coded is bad, but pya standalone has no exposed version information


def is_windows():
    if platform == "linux" or platform == "linux2":
        return False
    elif platform == "darwin":
        return False
    elif platform == "win32":
        return True
    else:
        raise ValueError('Unrecognized operating system: {}'.format(platform))



''' Spoof a whole bunch of stuff related to pya GUI.

    The problem this is trying to solve is when layout modules try to import GUI things and call them.
    When you run them in standalone, either those QElements don't exist, or they error for whatever reason.
    Nevertheless, you want to import that module to access its other functionality

    The desired behavior of these GUI things is pretty simple: don't error, do nothing.
    At the same time, we'd like the ability to override our override
    in case there are one or two important things, like instance.version() returning a string
'''

class NS_Catcher(type):
    ''' All this does is override the pya namespace with this class '''
    def __init__(cls, name, bases, dct):
        if pya is not None:
            setattr(pya, name, cls)
        super(NS_Catcher, cls).__init__(name, bases, dct)

    def __getattr__(cls, attr):
        return PhonyClass()


class PhonyClass(with_metaclass(NS_Catcher, object)):
    ''' It only ever gives instances of PhonyClass when called or as attributes.
        It is good for stifling those long chained calls like::

            pya.QFormBuilder().load(ui_file, pya.Application.instance().main_window()).findChild('ok').clicked(self.ok)

        That call will do nothing of course, but it also won't error.
    '''
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return PhonyClass()

    def __getattr__(self, attr):
        return PhonyClass()

    def __setattr__(self, attr, value):
        pass

    def __add__(self, other):
        return 0


def patch_environment():
    ''' Run this function, then your old scripts for klayout will work with the standalone.
        That's the idea anyways.

        Now we start changing things in the environment. Afterwards, this statement:
            import pya
        might give you GSI:pya or pymod:pya, depending.

        GUI and Application things are added and/or spoofed.
    '''
    if not isStandalone():
        sys.modules['pya'] = pya

    if not isGUI():
        class QMessageBox(PhonyClass): pass

        class QMessageBox_StandardButton(PhonyClass): pass

        class QFile(PhonyClass): pass

        class QIODevice(PhonyClass): pass

        class QFormBuilder(PhonyClass): pass

        class MessageBox(PhonyClass): pass

        class QDialog(PhonyClass): pass

        class Qt(PhonyClass): pass

        from lygadgets.technology import Technology
        pya.Technology = Technology
        Technology._load_salt()

    if not isGSI():

        class PhonyMainWindow(PhonyClass):
            initial_technology = ''

            def current_view(self):
                return None

        class PhonyInstance(PhonyClass):
            ''' This has to return a string sometimes '''
            def application_data_path(self):
                return klayout_home()

            def version(self):
                return klayout_version()

            def main_window(self):
                return PhonyMainWindow()

        class Application(PhonyClass):
            instance = PhonyInstance

        ''' Find the python modules that are present in klayout.
            These get lower import priority than system counterparts, if present
        '''
        if os.path.isdir(klayout_home()):
            for root, dirnames, filenames in os.walk(klayout_home(), followlinks=True):
                for dn in dirnames:
                    if dn in ['python', 'pymacros']:
                        sys.path.append(os.path.join(root, dn))





__version__ = '0.1.33'
__lygadget_link__ = ['future', 'xmltodict']

from lygadgets.environment import pya, isGUI, isGSI, patch_environment, klayout_home
from lygadgets.messaging import message, message_loud
from lygadgets.markup import xml_to_dict, lyp_to_layerlist
from lygadgets.system_linker import export_to_system
from lygadgets.cell_translation import anyCell_to_anyCell, any_read, any_write
from lygadgets.autolibrary import WrappedPCell, WrappedLibrary
from lygadgets.technology import Technology



''' Access the linkers from the command line
'''
import argparse
from lygadgets.salt_linker import link_any, unlink_any
from lygadgets import __version__


parser = argparse.ArgumentParser(description="lygadgets linkers between klayout and system namespaces")
parser.add_argument('sourcepackage', type=str,
                    help='the package to link. Can be a salt directory or a python directory, or the name of an installed python package')
parser.add_argument('-c', '--copy', action='store_true',
                    help='hard copy instead of symbolic linking')
parser.add_argument('-f', '--force', action='store_true',
                    help='overwrite anything present at the destination with the same name')
parser.add_argument('-v', '--version', action='version', version=f'%(prog)s v{__version__}')


def cm_link_any():
    args = parser.parse_args()
    the_links = link_any(args.sourcepackage, overwrite=args.force, hard_copy=args.copy)


unlink_parser = argparse.ArgumentParser(description='lygadgets unlink anything')
unlink_parser.add_argument('sourcepackage', type=str,
                    help='the package to unlink. Can be name of a package or technology')
unlink_parser.add_argument('-f', '--force', action='store_true',
                    help='delete non-symlink directories. Be careful')
unlink_parser.add_argument('-v', '--version', action='version', version=f'%(prog)s v{__version__}')

def cm_unlink_any():
    args = unlink_parser.parse_args()
    unlink_any(args.sourcepackage, force=args.force)



from setuptools import setup


def readme():
    with open('README.md', 'r') as fx:
      return fx.read()


setup(name='lygadgets',
      version='0.1.33',
      description='Tools to make klayout, the standalone, and python environments work better together',
      long_description=readme(),
      long_description_content_type='text/markdown',
      author='Alex Tait',
      author_email='alexander.tait@nist.gov',
      license='MIT',
      packages=['lygadgets'],
      install_requires=['future', 'xmltodict'],
      package_data={'': ['*.lym']},
      include_package_data=True,
      entry_points={'console_scripts': [
        'lygadgets_link=lygadgets.command_line:cm_link_any',
        'lygadgets_unlink=lygadgets.command_line:cm_unlink_any',
      ]},
      )



from __future__ import print_function
try:  # python 2 does not have FileNotFoundError. Check it
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

import os
import ast
import subprocess
from importlib import import_module
from shutil import rmtree, copytree
from types import ModuleType

from lygadgets.markup import xml_to_dict
from lygadgets.environment import klayout_home, is_windows


def symlink_windows(source, destination):
    # API based. The easy way
    try:
        os.symlink(source, destination)
        return
    except AttributeError:
        pass

    # Command line shell
    try:
        retval = subprocess.call(['ln', '-s', source, destination])
        assert retval == 0
        return
    except (subprocess.CalledProcessError, WindowsError, AssertionError):
        pass

    # Big magic with windows-specific package
    # From https://stackoverflow.com/questions/1447575/symlinks-on-windows
    try:
        import ctypes
        csl = ctypes.windll.kernel32.CreateSymbolicLinkW
        csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
        csl.restype = ctypes.c_ubyte
        flags = 0
        if source is not None and os.path.isdir(source):
            flags = 1
        if csl(destination, source, flags) == 0:
            raise ctypes.WinError('You do not have the permission to create a symbolic link.\nTry running as administrator')
        return
    except WindowsError:
        pass

    # Otherwise it didn't work
    raise WindowsError('Failed to find a way to link to a file in windows')


def lypackage_name(source):
    try:
        with open(os.path.join(source, 'grain.xml')) as grain:
            return xml_to_dict(grain.read())['salt-grain']['name']
    except FileNotFoundError as err:
        err.args = ((source +
                     ' does not appear to be a klayout package.' +
                     'It must have a grain.xml file.\n' +
                     str(err.args[0])), ) + err.args[1:]
        raise
    except KeyError as err:
        raise FileNotFoundError((source +
                                 ' : grain.xml does not define a "name" property.'))


def is_lypackage(source):
    if not os.path.isdir(source):
        return False
    try:
        lypackage_name(source)
        return True
    except (FileNotFoundError, KeyError):
        return False


def is_pypackage(source):
    if not os.path.isdir(source):
        return False
    return os.path.exists(os.path.join(source, '__init__.py'))


def is_pymodule(source):
    if not os.path.isfile(source):
        return False
    return os.path.splitext(source)[1] == '.py'


def is_lytech(source):
    ''' In this case the "source" is the full tech directory or the .lyt file
        It always returns the tech directory path
    '''
    if os.path.isfile(source):
        if source.endswith('.lyt'):
            source = os.path.dirname(source)
        else:
            return False
    else:
        for file_obj in os.listdir(source):
            if file_obj.endswith('.lyt'):
                return source
        else:
            return False


def is_lymacro(source):
    if not os.path.isfile(source):
        return False
    if not source.endswith('.lym'):
        return False
    return True


def is_pymacro(source):
    if not is_lymacro(source):
        return False
    with open(source) as macro:
        interpreter = xml_to_dict(macro.read())['klayout-macro']['interpreter']
    return interpreter.lower() == 'python'


def is_rubymacro(source):
    return is_lymacro(source) and not is_pymacro(source)


def module_from_str(module):
    if type(module) is ModuleType:
        return module
    elif type(module) is str:
        return import_module(module)
    else:
        raise TypeError('Argument must either be a module or a string')


def is_installed_python(module):
    try:
        module_from_str(module)
        return True
    except:
        return False


def srcdir_from_any(source):
    try:
        if os.path.exists(source):
            return os.path.realpath(source)
    except TypeError:
        pass
    if is_installed_python(source):
        module = module_from_str(source)
        # now figure out if it is a package or a non-packaged module
        try:
            return module.__path__[0]
        except (AttributeError, IndexError):
            return module.__file__
    else:
        raise FileNotFoundError('{} does not exist'.format(source))


def dest_from_srcdir(source, exclude_python_types=False):
    if is_lypackage(source):
        link_name = lypackage_name(source)
        link_dir = os.path.join(klayout_home(), 'salt')
    elif is_lytech(source) != False:
        enclosing_dir = is_lytech(source)
        link_dir = os.path.join(klayout_home(), 'tech')
        link_name = os.path.splitext(os.path.basename(enclosing_dir))[0]
    elif is_lymacro(source):
        link_dir = os.path.join(klayout_home(), 'pymacros' if is_pymacro(source) else 'macros')
        link_name = os.path.basename(source)
    elif is_pypackage(source) or is_pymodule(source):
        if not exclude_python_types:
            link_dir = os.path.join(klayout_home(), 'python')
            link_name = os.path.basename(source)
        else:
            # return None
            raise TypeError('Python code found but it is being excluded: ' + source)
    else:
        raise TypeError('Did not recognize the klayout relevance of {}'.format(source))

    if not os.path.exists(link_dir):
        os.mkdir(link_dir)
    link = os.path.join(link_dir, link_name)
    return link


def link_any(any_source, overwrite=False, hard_copy=False, keep_links=False, exclude_python_types=False):
    ''' Directories take precedence over installed python module

        Platform independent.

        If keep_links=False, overwrites existing symbolic links.

        Returns the full paths of source and destination if the link was created, otherwise None for both.

        If you have given it a python package or an installed module with .lym's in it, klayout will automatically find them.
        Note to make sure they come along with the pip distro, use "package_data" in setup.py
    '''
    src = srcdir_from_any(any_source)
    try:
        dest = dest_from_srcdir(src, exclude_python_types=exclude_python_types)
    except TypeError as err:
        print(err)
        return None, None

    if src == dest:
        # Prevent circular reference
        print('Circular reference')
        print(src)
        return None, None
    if os.path.islink(dest):
        if not keep_links:
            os.remove(dest)
        else:
            print(any_source, 'already installed')
            return None, None
    if os.path.exists(dest):
        if overwrite:
            rmtree(dest)
        else:
            print('Not linking. Destination has a non-symlink item present that would be overwritten.')
            print('Use -f command line option to overwrite, or use lygadgets_unlink to remove it first')
            print(dest)
            return None, None

    if hard_copy:
        copytree(src, dest)
        print('Successfully created a hard copy')
    else:
        if not is_windows():
            os.symlink(src, dest)
        else:
            symlink_windows(src, dest)
        print('Successfully created a symbolic link')
    print('From:', src)
    print('To:  ', dest)

    # __lygadget_link__ is the special top package attribute that triggers more linking.
    # A list of strings/modules that must be installed or discoverable by import_module
    others_to_link = []
    if is_installed_python(any_source):
        module = module_from_str(any_source)
        try:
            others_to_link = module.__lygadget_link__
        except AttributeError:
            pass
    elif is_pypackage(any_source) or is_pymodule(any_source):
        # Find that assignment, but don't import the module
        if is_pypackage(any_source):
            code_file = os.path.join(any_source, '__init__.py')
        else:
            code_file = any_source
        with open(code_file) as fx:
            code_text = fx.read()

        for line in code_text.split('\n'):
            if '__lygadget_link__' in line:
                code_ast = ast.parse(line)
                try:
                    assign_ast = code_ast.body[0]
                    assert isinstance(assign_ast, ast.Assign)
                    assert assign_ast.targets[0].id == '__lygadget_link__'
                    if isinstance(assign_ast.value, ast.Constant):
                        others_to_link = [assign_ast.value.value]
                    elif isinstance(assign_ast.value, (ast.List, ast.Tuple)):
                        others_to_link = [e.value for e in assign_ast.value.elts]
                    else:
                        raise TypeError('__lygadget_link__ needs to be found in simple assignment')
                except (AttributeError, TypeError, AssertionError):
                    # something went wrong
                    print('Something went wrong linking lygadget dependencies in this line')
                    print(line)
                break

    # Do dependency linking
    for other in others_to_link:
        subsrc, subdest = link_any(other, keep_links=True)
        if subsrc is not None:
            print('Dependency linked:', any_source, '->', other)

    return src, dest


def unlink_any(installed_name, force=False):
    matches = []
    search_dir = os.path.join(klayout_home(), 'python')
    for fname in os.listdir(search_dir):
        if fname == installed_name:
            matches.append(os.path.join(search_dir, fname))
    search_dir = os.path.join(klayout_home(), 'tech')
    for fname in os.listdir(search_dir):
        if fname == installed_name:
            matches.append(os.path.join(search_dir, fname))
    if len(matches) == 0:
        print('Did not find matching installed package for "{}"'.format(installed_name))
    elif len(matches) > 1:
        print('Multiple matches found. Delete manually')
        print('\n'.join(matches))
    else:
        match = matches[0]
        if os.path.islink(match):
            os.remove(match)
            print('Removed symlink', match)
        elif force:
            rmtree(match)
            print('Removed directory', match)
        else:
            print(match, 'is a directory, not a symlink.')
            print('Use the -f option if you are sure you want to permanently delete.')
